import functools
import warnings

from six import string_types


class _DeprecatedFunction(object):

    def __init__(self, func, message, obj=None, objtype=None):
        super(_DeprecatedFunction, self).__init__()
        self._func = func
        self._message = message
        self._obj = obj
        self._objtype = objtype

    def _get_underlying_func(self):
        returned = self._func
        if isinstance(returned, classmethod):
            if hasattr(returned, '__func__'):
                returned = returned.__func__
            else:
                returned = returned.__get__(self._objtype).__func__
        return returned

    def __call__(self, *args, **kwargs):
        func = self._get_underlying_func()
        warning = "{0} is deprecated.".format(self._get_func_str())
        if self._message is not None:
            warning += " {0}".format(self._message)
        warnings.warn(warning, DeprecationWarning, stacklevel=2)
        if self._obj is not None:
            return func(self._obj, *args, **kwargs)
        elif self._objtype is not None:
            return func(self._objtype, *args, **kwargs)
        return func(*args, **kwargs)

    def _get_func_str(self):
        func = self._get_underlying_func()
        if self._objtype is not None:
            return '{0}.{1}'.format(self._objtype.__name__, func.__name__)
        return '{0}.{1}'.format(func.__module__, func.__name__)

    def __get__(self, obj, objtype):
        return self.bound_to(obj, objtype)

    def bound_to(self, obj, objtype):
        return _DeprecatedFunction(self._func, self._message, obj=obj,
                                   objtype=objtype)

    @property
    def __name__(self):
        return self._get_underlying_func().__name__

    @property
    def __doc__(self):
        returned = self._get_underlying_func().__doc__
        if returned:  # pylint: disable=no-member
            returned += "\n.. deprecated\n"  # pylint: disable=no-member
            if self._message:
                returned += "   {0}".format(
                    self._message)  # pylint: disable=no-member
        return returned

    @__doc__.setter
    def __doc__(self, doc):
        self._get_underlying_func().__doc__ = doc


def deprecated(func=None, message=None):
    """Marks the specified function as deprecated, and emits a warning when
    it's called.

    >>> @deprecated(message='No longer supported')
    ... def deprecated_func():
    ...     pass

    This will cause a warning log to be emitted when the function gets called,
    with the correct filename/lineno.

    .. versionadded:: 0.12
    """
    if isinstance(func, string_types):
        assert message is None
        message = func
        func = None

    if func is None:
        return functools.partial(deprecated, message=message)

    return _DeprecatedFunction(func, message)



import functools
from contextlib import contextmanager

import pytest

import vintage

_EXPECTED_VALUE = 6


def test_vintage(deprecated, expected_message):
    with _assert_single_deprecation() as w:
        deprecated()
    assert w.message == expected_message


def test_deprecated_property(message):

    class Sample(object):

        @property
        @vintage.deprecated(message=message)
        def prop(self):
            return _EXPECTED_VALUE

    with _assert_single_deprecation():
        assert Sample().prop == _EXPECTED_VALUE

def test_deprecated_doc(deprecated):
    assert '.. deprecated' in [l.strip() for l in deprecated.func.__doc__.splitlines()]


##########################################################################

@pytest.fixture
def expected_message(deprecated, message, deprecated_type):
    if deprecated_type == 'method':
        func = deprecated.func._func  # pylint: disable=protected-access
        name = 'Sample.func'
    else:
        func = deprecated.func
        name = '{}.func'.format(__name__)
    assert func.__name__ == 'func'
    returned = '{} is deprecated.'.format(name)
    if message:
        returned += ' {}'.format(message)
    return returned

@pytest.fixture(params=['method', 'function'])
def deprecated_type(request):
    return request.param


@pytest.fixture
def deprecated(deprecated_type, message):
    if deprecated_type == 'method':
        class Sample(object):

            @_deprecate(message)
            def func(self, a, b, c):
                "docstring"
                return a + b + c

        returned = Sample().func
    elif deprecated_type == 'function':
        @_deprecate(message)
        def func(a, b, c):
            "docstring"
            return a + b + c
        returned = func

    else:
        raise NotImplementedError()  # pragma: no cover

    return functools.partial(returned, 1, 2, _EXPECTED_VALUE - 1 - 2)


def _deprecate(message):
    """Helper to sometimes simulate calling ``deprecated`` with parameters and sometimes plainly without any arguments
    """
    def decorator(func):
        if message:
            return vintage.deprecated(message=message)(func)
        return vintage.deprecated(func)
    return decorator


@pytest.fixture(params=[None, 'message here'])
def message(request):
    return request.param


@contextmanager
def _assert_single_deprecation():
    with pytest.deprecated_call() as w:
        yield _RecordProxy(w)
    assert len(w) == 1, 'No single warning captured'


class _RecordProxy(object):

    def __init__(self, record):
        self._record = record

    def __getattr__(self, attr):
        # record.message is actually our warning
        w = self._record[0].message
        if attr == 'message':
            assert len(w.args) == 1
            return w.args[0]
        return getattr(w, attr)
